
/* Twiddler Driver for Linux V2.0
 * v2.0: 7/20/98
 * 8/14/95
 * written by:Jeff Levine
 * mouse support added 1/15/96
 * 
*/

#include "twiddler.h"

int twidfd,ttyfd,fifo; /* globally known for signal catcher to close */
char **table; /* everybody needs to know these two */
FILE *a2x_file = NULL; 
int X_VT, current_vt;
struct state states;
minfo m_info;

#ifdef XSTUFF
Display *dpy=NULL;
#endif



main (int argc, char * argv[])
{

  states.flags = 0; 
  if (signal(SIGPIPE,X_off)==SIG_ERR)
    { 
      printf ("Can't get signal SIGPIPE \n");
      exit (-1);
    }
  if (signal(SIGINT,quit)==SIG_ERR)
    { 
      printf ("Can't get signal SIGINT \n");
      exit (-1);
    }
  /* let's catch some other signals here to close properly */
  if (signal(SIGTERM,quit)==SIG_ERR)
    {
      printf ("Can't get signal SIGTERM \n");
      exit (-1);
    }
  /* initialize state, set defaults, and open ports... all that stuff */
  init (&states,argc,argv);
  unlink (TWID_TEMP);
  /* now we'll enter the read/write loop using 5 byte blocks */
  mainloop ();
}

void mainloop ()
{
  int i,j;
  char port [2],text[MAXLINE],*filename;
  unsigned char byte[1],buffer[5];
#ifdef linux
  struct vt_stat vst;
#endif
  FILE *foo;

  while (1)
    {
#ifdef linux
      ttyfd = open ("/dev/console", O_WRONLY); 
      
      /* this makes sure we are current*/
      ioctl (ttyfd,VT_GETSTATE, &vst);
      close (ttyfd);
      
      if (current_vt != vst.v_active)
	{
	  current_vt = vst.v_active;
	  if (current_vt == X_VT)
	    {
#else
	      
      if (a2x_file == (FILE *) NULL)
	{

#endif
	      m_info.xterm_on = 1;
	      a2x_file = popen  ("a2x -a","w"); 
	      /*	      fprintf (stderr,
		       "popen = %p\n",a2x_file);*/
	      if (!a2x_file)
		    {
		      fprintf (stderr,
		       "Could not Open a2x!! X support will not work!\n");
#ifdef linux
		      fprintf (stderr,"defaulting to console mode\n");
#else
		      fprintf (stderr,"exiting\n");
		      quit(1);
#endif

#ifdef XSTUFF
		      dpy=NULL;
#endif		     
		      m_info.xterm_on = 0;
		    }
#ifdef linux
	    }
	  else
	    {
#  ifdef XSTUFF
	      dpy=NULL;
#  endif		     
	      m_info.xterm_on = 0;
	      if (a2x_file)
		pclose (a2x_file);
	      a2x_file = NULL;
	    }
	}
#else
	}
#endif

      if (i=open("/tmp/twid_signal",O_RDONLY)!=-1)
	{
	  reinit (1);
	  close (i);
	  unlink ("/tmp/twid_signal");
	}
      for (i=0;i<5;i++)
	{
	  while (read (twidfd, buffer+i, 1)<=0);
	  
	  if ((i==0)&&(buffer[0]&0x80))
	    {
	      i--; /* we're out of sync... let's keep trying */
#ifdef DEBUG
	      printf ("Out of sync!\n"); 
#endif
	    }
	  if ((i>0) && (!buffer[i]&0x80))
	    {
	      buffer[0]= buffer[i];
	      i=1;
#ifdef DEBUG
	      printf ("header byte in the middle of a sequence\n");
#endif
	    }    /* Found a header byte in the middle of a sequence! */
	  /* start over with that byte first */ 
	}
#ifdef DEBUG
      printf ("Buffer Contents: ");
      for (i=0;i<5;i++)
	printf ("%03x ",buffer[i]);
      printf ("\n");
#endif 
      if (processbuffer (buffer,&states))
	/* if there has been a change in state, or we are in repeat
	   mode, then we will write the buffer contents */
	{
#ifdef linux
	  if (m_info.xterm_on) 
	    {
#endif
	      get_text (&states,text);
	      if (a2x_file) fprintf (a2x_file,"%s",text);
	      if (a2x_file) fflush (a2x_file);
	      /* write to a2x_file with a2x modifiers */
#ifdef linux
	    }
	  else
	    {
	      get_text (&states,text);
	      ttyfd = open ("/dev/console",O_WRONLY);
	      for (i=0; i<strlen(text); i++)
		ioctl (ttyfd,TIOCSTI,text+i);
	      close (ttyfd);
	      /* I believe that the above line is Linux specific..
		 any porting would have to take this into account */
	    }
#endif
	}
    }
}

int processbuffer (unsigned char *buffer, STATE events)
{
  
  /* returns 1 if new event has been added, 0 otherwise */
  
  /* event enabled upon key press
     event generated upon key release
     unless repeat (then it waits for a time and the generates an event */
  int index=0;
  char CAPS_CODE [8]={27,27,27,27,27,0};
  char NUM_CODE [8]={27,27,27,27,27,1,0};  
  char SCROLL_CODE [8]={27,27,27,27,27,2,0};
  char X_TOGGLE [8]={27,27,27,27,27,3,0}; 
  /* with vt switching I think his is obsolete now */
  char NULL_CODE [8]={27,27,27,27,27,4,0};
  char VT_CODE [8]={27,27,27,27,27,5};
  char CTALDEL_CODE [8]={27,27,27,27,27,6};
  char mouse_but, *temp;
  static int mouse_but_state = M_OFF;
  static int lastvert = 0;
  static int lasthoriz = 0;
  float delta;
  int meancx,meancy, angx, angy;
  int dx,dy,cx,cy;
  int x_res, y_res,temp_res,fil_lev;
  index |= ((buffer[0]&0x7f) | ((buffer[1]&0x3f)<<7));
  
  events->last_state = events->cur_state;
  events->cur_state.index = index;
  
  if (events->cur_state.index>events->last_state.index)
    (events->flags |= EVENT_ENABLE);
  
  if (events->time_count<REPEAT_TIME_THRESH)
    events->time_count++;
  if (m_info.xterm_on)
    {
      /* handle mouse togggle */
      mouse_but = buffer[1]&0x40;
      events->cur_state.horiz_tilt = 0 | (((buffer[3]&0x7c)>>2) | 
					  ((buffer[4]&0x07)<<5));
      events->cur_state.vert_tilt = 0 | ((buffer[2]&0x7f) | 
					 ((buffer[3]&0x01)<<7));
      if (!(buffer[3]&0x02))
	events->cur_state.vert_tilt += 256;
      if (!(buffer[4]&0x08))
	events->cur_state.horiz_tilt += 256;
      /* flip the high order bit to get offset binary,
	 then just use integer subtract... etc.. */
      
      /* move mouse if need be. */

      cx = -(events->cur_state.horiz_tilt - 256);
      cy = -(events->cur_state.vert_tilt - 256);
      if (m_info.mouse_reverse) cy=-cy;
      /* ok, now store raw number, get raw-ave filter,  */
      /* then linearize that # */
      
      getmean (m_info.mouse_filter, &meancx,&meancy,cx,cy,
	       m_info.FILTER_LEVEL, m_info.rawmean);      
	
      if (meancx<0)
	angx = -LIN_ANGLE [-meancx];
      else
	angx = LIN_ANGLE [meancx];
      if (meancy<0)
	angy = -LIN_ANGLE [-meancy];
      else
	angy = LIN_ANGLE [meancy];

      dx = angx - lasthoriz;
      dy = angy - lastvert;
      m_info.mouse_filter = add_to_filter (m_info.mouse_filter, dx,dy,cx,cy,
				    m_info.FILTER_LEVEL, m_info.mouse_on);  
      lasthoriz = angx;
      lastvert = angy;

      /* for now -- variable filter length I'll try later */
      /* raise mouse buttons if need be */
      if (m_info.mouse_on && mouse_but==0)
	{
	  clear_mouse_state (0);
	  m_info.mouse_on = 0;
	}

      if ((index&0x03)==0)
	{	
	  /* mouse buttons go back to up state. */
	  if (a2x_file) fprintf (a2x_file,"%c%c0%c",20,2,20);
	  if (a2x_file) fflush (a2x_file);
	}

      if (mouse_but)
	{
	  handle_mouse_activity (index, events); 
	  return (0);
	}
      else
	{
	  if (a2x_file) fprintf (a2x_file,"%c%c0%c",20,2,20);
	  if (a2x_file) fflush (a2x_file);
	}

    }

  if (!(events->flags&EVENT_ENABLE))
    return (0); /* if key events not enabled, let's return 0 */
    
  /* if the last state was a mapped character and we just released */
  if (events->cur_state.index < events->last_state.index)
    {
      temp = search_table(events->last_state.index);
      /* first let's check for special cases */
      if (temp)
	{
	  if (!strcmp (temp,CAPS_CODE))
	    {
	      events->flags &= ~EVENT_ENABLE;
	      if ((events->flags)&CAPS_LOCK)
		events->flags &= ~(CAPS_LOCK);
	      else
		events->flags |= CAPS_LOCK;
	      return (0); /* not really an event */
	    }
	  if (!strncmp (temp, VT_CODE,6))
	    {	      
#ifdef linux
	      int term = temp[6];
	      ttyfd = open ("/dev/console",O_RDWR);
	      ioctl (ttyfd, VT_ACTIVATE, term);	   
	      close (ttyfd);
	      /** LINUX SPECIFIC VT SWITCHING **/
#endif
	      return (0);
	    }
	  if (!strncmp (temp,CTALDEL_CODE,6))
	    {	      
	      char *shutdown_command=temp+6;
	      system (shutdown_command); 
	      return (0);
	    }
	  if (!strcmp (temp,NUM_CODE))
	    {
	      events->flags &= ~EVENT_ENABLE;
	      if (events->flags&NUM_LOCK)
		events->flags &= ~(NUM_LOCK);
	      else
		events->flags |= NUM_LOCK;
	      return (0); /* not really an event */
	    }      
	  if (!strcmp (temp,X_TOGGLE))
	    {
	      events->flags &= ~EVENT_ENABLE;
	      X_VT = current_vt;
	      return (0); /* not really an event */
	    }
	  if (!strcmp (temp,NULL_CODE))
	    {
	      events->flags &= ~EVENT_ENABLE;
	      if (m_info.xterm_on) 
		{
		  if (a2x_file) fprintf (a2x_file,"%c",0);
		  if (a2x_file) fflush (a2x_file);
		}
	      else
		{
		  char null=0;
		  ttyfd = open ("/dev/console",O_WRONLY);
		  ioctl (ttyfd,TIOCSTI,&null);
		  close (ttyfd);
		} 
	      return (0); /* not really an event */
	    }
	  if (!strcmp (temp,SCROLL_CODE))
	    {
	      events->flags &= ~EVENT_ENABLE;
	      if (events->flags&SCROLL_LOCK)
		events->flags &= ~(SCROLL_LOCK);
	      else
		events->flags |= SCROLL_LOCK;
	      events->last_event=events->last_state;
	      if (events->flags&SCROLL_LOCK)
		events->last_event.index=SCROLL_ON_INDEX;
	      else
		events->last_event.index=SCROLL_OFF_INDEX;
	      return (1); /* scroll lock produces an event */
	    }
	}
      /* generate event based on the last state */
      if (events->repeat_count>=REPEAT_THRESH)/* we just repeated some stuff */
	events->flags &= ~REPEAT_ENABLE;
      else
	events->flags |= REPEAT_ENABLE; /* else leave repeat option open */
      events->time_count = 0;
      events ->repeat_count=0; 
      events->last_event=events->last_state;
      events->flags &= ~EVENT_ENABLE;
      return (1);
    }

  /* we just pressed something, lets check if we're ready
     to go into repeat mode */

  if ((events->flags&REPEAT_ENABLE) && 
      (events->time_count-events->repeat_count<REPEAT_TIME_THRESH) &&
      (search_table (events->cur_state.index)) &&
      (events->cur_state.index==events->last_event.index))
    events->repeat_count++; /* if state is a repeat state, and not first instance */
  else events->repeat_count=0;
  if (events->repeat_count>REPEAT_THRESH)
    events->repeat_count=REPEAT_THRESH; 
  /*prevents uncontrolled growing
    if we fall asleep on the space bar (so to speak :) */
  
  if (events->repeat_count == REPEAT_THRESH)
    {
      /* generate an event based on the current state */
      events->last_event=events->cur_state;
      return(1);
    }
  return (0);
}

struct dxdy *setup_filter (int filter_level)
{
  int i;  
  struct dxdy *temp;
  struct dxdy *orig;
  orig = calloc (1,sizeof (struct dxdy));
  temp=orig;
  for (i=0; i<(filter_level-1); i++)
    {
      temp->next = calloc (1,sizeof (struct dxdy));
      temp->next->prev = temp;
      temp = temp->next;
    }
  temp->next = orig;
  orig->prev=temp;
  return orig;
}

void kill_filter (struct dxdy *filter, int filter_level)
{
  int i;
  for (i=0; i<filter_level;i++)
    {
      filter->dx=0;
      filter->dy=0;
      filter=filter->next;
    }
}

void getmean (struct dxdy *filter, int *cx, int *cy, int newcx, int newcy,
	      int filter_size, int numavg)
{
  int i;
  *cx = 0;
  *cy = 0;

  if (numavg>filter_size)
    numavg=filter_size;
  if (numavg<1) numavg =1;

  for (i=0;i<(numavg-1); i++)
    {
      *cx += filter->cxraw;
      *cy += filter->cyraw;
      filter= filter->prev;
    }
  *cx += newcx; /* current values are the last to be added */
  *cy += newcy;
  *cx /= numavg;
  *cy /= numavg;
}

void clear_mouse_state (int clearraw)
{
  int i;

  m_info.old_den = 1; /* just don't make it 0 */
  m_info.dx_rem = 0;
  m_info.dy_rem = 0; /* throw out remainder */
  for (i=0;i<m_info.FILTER_LEVEL;i++)
    {
      /* empty filter */
      m_info.mouse_filter->dx = 0;
      m_info.mouse_filter->dy = 0;
      if (clearraw)
	{
	  m_info.mouse_filter->cxraw = 0;
	  m_info.mouse_filter->cyraw = 0;
	}
      m_info.mouse_filter = m_info.mouse_filter->next;
    }
}


struct dxdy *add_to_filter (struct dxdy *filter, int dx, int dy,
			    int cx,int cy,
			    int filsize, int add_del)
{
  filter = filter->next;
  if (add_del)
    {
      filter->dx = dx;
      filter->dy = dy;
    }
  else
    {
      filter->dx = 0;
      filter->dy = 0;
    }
  filter->cxraw = cx;
  filter->cyraw = cy;
  return (filter);
}  
#ifdef OLD_SHIT
struct dxdy* filter_mouse (struct dxdy *filter, int filter_level,
			   int res_level, int *dx, int * dy)
{
  int i;
  
  if (res_level==0)
    {
      *dx= *dx *2;
      *dy= *dy *3;
    }
  if (res_level==1)
    {
      *dx= *dx *3;
      *dy= *dy *5;
    }
  filter->dx = *dx;
  filter->dy = *dy;
  

  /* add dynamic filtering and 
     dynamic resolution */
  *dx = 0;
  *dy = 0;
  for (i=0;i<(filter_level-1); i++)
      filter=filter->prev;
  for (i=0;i<filter_level; i++)
    {
      *dx += filter->dx;
      *dy += filter->dy;
      filter=filter->next;
    }
  *dx /= filter_level;
  *dy /= filter_level;
  return filter;
}
#endif
void get_text (STATE events,char *text)
{
  char line2[MAXLINE]="",line[MAXLINE]="", *temp;
  int i;
  
  *text=0;
  temp = search_table(events->last_event.index);
  if (temp==NULL)
    {
      if (m_info.xterm_on)  /* in X if we haven't found an entry.. try
			appending with the correct modifier */
	{
	  /* a2x will accept ctrl, alt, and shift modifiers, so if we have nothing
	     else defined, we can put these modifiers in our text*/
	  
	  if (events->last_event.index &(CTRL_INDEX))
	    sprintf  (line,"%s%c%c",line,20,3);
	  if (events->last_event.index &(ALT_INDEX))
	    sprintf (line,"%s%c%c",line,20,13);
	  if (events->last_event.index &(SHIFT_INDEX))
	    sprintf (line,"%s%c%c",line,20,19);
	  temp =search_table(events->last_event.index&0xff);
	  if (temp)
	    {
	      strcat (line,temp);
	      strcpy (text,line);
	    }
	}
      return;
    }
  strcpy (line,temp);
  if (events->flags&CAPS_LOCK)
    {
      for (i=0;i<strlen(line);i++)
	if ((line[i]>='a') && (line[i]<='z'))
	  line[i] = line[i]-('a'-'A');
    }
  if (events->flags&NUM_LOCK)
    {
      if (search_table(events->last_event.index |(NUM_INDEX)))
	{
	  strcpy (line2,search_table(events->last_event.index |(NUM_INDEX)));
	  if ((strlen(line2)==1) && /* works only for single characters */
	      ((*line2>='0') && (*line2<='9')))
	    strcpy (line,line2);
	}
    }
  /* scroll lock has no instantanious function... */
  strcpy (text,line);
  return;
}

void free_table()
{
  int i;
  for (i=0;i<8192;i++)
  {
    if (table[i])
      free(table[i]);
  }
  free(table);
}

void init_table()
{
  table = (char **) calloc (8192,sizeof(char **));
}

char *search_table  (int super_index)
{
  return table[super_index];
} 

#if defined(__alpha) || defined(linux)
#  include <pty.h>
#endif

int open_serial (char *device)
     /* opens serial port to 2400,8N1 */
{
  struct termios term;
  int fd;
 int mybit;
#ifdef DEBUG  
  int line_state;
  printf ("Opening %s...\n",device);
#endif

  if ((fd = open (device , O_RDONLY))<0)
    {
      perror("open serial device");
      return (-1);
    }
  
  if (!isatty(fd))
    {
      printf ("This is not a TTY device!\n");
      return (-1);
    }
  
  if ((tcgetattr (fd ,&term))<0)
    {
      printf ("%d - ",errno);
      printf ("Couldn't get tty info!\n");
      return (-1);
    }

#ifdef __alpha
  cfsetospeed (&term,2400);
  cfsetispeed (&term,2400);
#else
  cfsetospeed (&term,B2400);
  cfsetispeed (&term,B2400);
#endif

  if (tcsetattr (fd,TCSAFLUSH, &term) < 0)
    perror ("tcsetattr/TCSAFLUSH");
  if (tcflush (fd, TCIOFLUSH) < 0)
    perror ("tcflush/TCIOFLUSH");
  
  if ((tcgetattr (fd ,&term))<0)
    {
      printf ("%d - ",errno);
      printf ("Couldn't get tty info!\n");
      return (-1);
    }
  /* put in raw data mode */
  term.c_lflag &= ~(ICANON | ECHO | IEXTEN | ISIG);
  term.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
  term.c_cflag &= ~(CSIZE | PARENB);
  term.c_oflag &= ~OPOST;
  term.c_cflag |= CS8; /* set to 8 stop bits */
  term.c_cflag |= CLOCAL;
  term.c_cc[VMIN]=1;
  term.c_cc[VTIME]=0;

  if(tcsetattr (fd,TCSAFLUSH, &term) < 0)
    perror ("tcsetattr/TCSAFLUSH");
  if(tcflush (fd, TCIOFLUSH) < 0)
    perror ("tcflush/TCIOFLUSH");
  if(tcflow (fd,TCOOFF) < 0)
    perror ("tcflow/TCOOFF");

  mybit = TIOCM_DTR;
  ioctl (fd, TIOCMBIC, &mybit);

#ifdef DEBUG  
  ioctl (fd, TIOCMGET, &line_state);
  printf ("Line State: ");
  if (line_state & TIOCM_LE ) printf ("LE  "); else printf ("!LE  ");
  if (line_state & TIOCM_DTR) printf ("DTR "); else printf ("!DTR ");
  if (line_state & TIOCM_RTS) printf ("RTS "); else printf ("!RTS ");
  if (line_state & TIOCM_ST ) printf ("ST  "); else printf ("!ST  ");
  if (line_state & TIOCM_SR ) printf ("SR  "); else printf ("!SR  ");
  if (line_state & TIOCM_CTS) printf ("CTS "); else printf ("!CTS ");
  if (line_state & TIOCM_CAR) printf ("CAR "); else printf ("!CAR ");
  if (line_state & TIOCM_RNG) printf ("RNG "); else printf ("!RNG ");
  if (line_state & TIOCM_DSR) printf ("DSR "); else printf ("!DSR ");
  printf ("\n");
#endif

  return (fd);
}

int init(STATE states,int argc,char *argv[])
{
  char device[10]=DEFAULT_DEVICE,line[MAXLINE];
  int opt;
  FILE *pidfile;
  int i,pid,initialize=0;
  char *filename;

  /* let's see if we should even try */
  
  if (setuid(0)<0) root_error();
  
  if (!(filename = getenv("TWID_DEFAULTS")))
    {
      filename = calloc (25,sizeof (char));
      strcpy (filename , DEFAULT_FILE);
    }
  init_table();

  m_info.FILTER_LEVEL = 7;
  m_info.old_den = 1;
  m_info.dx_rem = 0;
  m_info.dy_rem = 0;
  m_info.mouse_res=1;
  m_info.rawmean = 3;
  m_info.m_hr_ind = 0x20;
  m_info.m_xhr_ind = 0x30;
  m_info.m_rev_ind = 0x54;
  m_info.m_lb_ind = 0x01;
  m_info.m_mb_ind = 0x02;
  m_info.m_dc_ind = 0x0c;
  m_info.m_rb_ind = 0x03;
  m_info.high_res = 15;
  m_info.mid_res = 30;
  m_info.norm_res = 60;
  m_info.tilt_constant = 0;
  m_info.tilt_divisor = 0;
  m_info.filmin = 4;
  m_info.delmax = 50;
  m_info.accel_thresh = 0;
  m_info.accel_factor = 0;
  m_info.sensitivity = 5;
  m_info.m_samplesize = 0;
  m_info.m_samplenum = 0;
  m_info.saved_dx = 0;
  m_info.saved_dy = 0;

  X_VT = 7; 
  /* let's see what the user has in store for us */
  while ((opt=getopt(argc,argv,CLI_OPTIONS))!=-1)
    {
      switch (opt)
	{   
	case 'k':
	  if (getuid()>0)
	    root_error();
	  if (pidfile = fopen (TWID_FILE,"r"))
	    {
	      fscanf(pidfile,"%d",&pid);
	      if (kill (pid,SIGTERM)<0)
		{
		  printf ("Couldn't kill Twiddler!\n");
		  exit (-1);
		}	 
	      fclose (pidfile);
	    }
	  exit (-1);
	case 'x':
	  X_VT = atoi (optarg);
	  break;
	case 'p':
	  if (getuid()>0)
	    root_error();
#ifdef linux
	  strcpy (device,"/dev/cua");
#endif
#ifdef __alpha
	  strcpy (device,"/dev/tty0");
#endif
#ifdef __sgi
	  strcpy (device,"/dev/ttyd");
#endif
	  strncat(device,optarg,1);
	  break;
	case 'f':
	  if (!unique())
	    {
	      if (pidfile = fopen (TWID_FILE,"r"))
		{
		  fscanf(pidfile,"%d",&pid);
		  fclose (pidfile);
		  pidfile = fopen (TWID_TEMP,"w"); /* actually a temp file*/
		  fprintf (pidfile,"%s",optarg);
		  fclose(pidfile);
		  system("touch /tmp/twid_signal");
		} 
	      exit (1);
	    }
	  else
	    {
	      /*  this is it! let's just get it */
	      if (read_file(filename)==-1)
		{
		  printf ("Twiddler Keymap: Error reading default file %s !!\n",filename);
		  printf ("Check that file exists and try again\n");
		  exit (-1);
		}
	      if (read_file(optarg)==-1)
		{
		  fprintf (stderr,"Twiddler Keymap: Error reading file %s!\n",optarg);
		  fprintf (stderr,"Additions will not be active...\n");
		}

	      initialize=1;
	    }
	  break;
	default:
	  printf ("Invalid argument... %c\n",opt);
	  printf ("Usage: %s [-p #] [-k] [-f filename] [-x #]\n", argv[0]);
#ifdef linux
	  printf ("Where -p # is a number from 0 to 3\n");
	  printf ("corresponding to cua0-cua3 (COM1-COM4 in DOS)\n");
#endif
#ifdef __alpha
	  printf ("Where -p # is a number from 0 to 1\n");
	  printf ("corresponding to tty00-tty01\n");
#endif
#ifdef __sgi
	  printf ("Where -p # is a number from 1 to 2\n");
	  printf ("corresponding to ttym1-ttym2\n");
#endif
	  printf ("-x # is a number from 1 to MAX_NR_CONSOLES\n");
	  printf ("-k will kill the driver completely\n");
	  printf ("Note: you must be root to execute any option besides -f\n");
	  printf ("Calling with no arguments resets the defaults\n");
	  quit (SIGQUIT);     
	};
    }
  if (!initialize)
    if (unique())
      {
	if (read_file(filename)==-1)
	  {
	    printf ("Twiddler Keymap: Error reading default file %s !!\n",filename);
	    printf ("Check that file exists and try again\n");
	    exit (-1);
	  }
      }
    else
      {
	if (pidfile = fopen (TWID_FILE,"r"))
	  {
	    fscanf(pidfile,"%d",&pid);
	    fclose (pidfile);
	    pidfile = fopen (TWID_TEMP,"w");
	    fprintf (pidfile ,"%s",filename);
	    fclose(pidfile);
	    system ("touch /tmp/twid_signal"); 
	    /* send signal to main program 
	       alerting it that we want to change
	       defaults*/
	  }	
	exit (1);
      }

#ifdef DEBUG2
  {
    int i,j;
    for (i=0;i<32;i++)
      {
	if (table_size[i])
	  {
	    printf ("BUCKET %d SIZE: %d\n",i,table_size[i]);
	    for (j=0;j<table_size[i];j++)
	      printf ("Bucket: %d Elem: %d Map: %s\n",i,table[i][j]->index,table[i][j]->str);
	  }
	else printf ("Bucket %d empty\n",i);
      }
    quit(SIGTERM); /*lets not process, just print out table */
    exit (1);
  }
#endif
  /* OK, now here we are unique and root */
#ifndef GDB  /* don't fork under debugger */
  switch (fork())
    {
    case -1: 
      printf ("Fork Error... Exiting\n");
      exit(-1);
    case 0: 
      break;
    default: exit(0);
    }
  if (setsid()<0) 
    {
      printf ("setsid error.. Exiting\n");
      exit (-1);
    }
  umask (0); /* just make sure we don't have anything funny */
#endif
  /* by now we've forked and killed papa, lets fire it up */

  /* inital states */
  states->cur_state.index=0;
  states->last_event.index=-1;
  states->last_state.index=0;

  /* OK, we're running for good, let's lock */
  /* yes this is not a great locking system or anything */
  /* but it keeps the casual user from duplicating by accident */

  unlink (TWID_FILE);
  if (!(pidfile = fopen (TWID_FILE,"w")))
    {
      printf ("Couldn't create file %s Exiting.. \n ",TWID_FILE);
      quit (SIGQUIT);
    }
  fprintf (pidfile,"%d",getpid());
  fclose (pidfile);
  /* time to open the port */
  m_info.mouse_filter = setup_filter (m_info.FILTER_LEVEL);
  twidfd = open_serial (device);
  if (twidfd<0)
    {
      printf ("Couldn't open Serial Port %s!\n",device);
      quit (1);
    }
  /* check for twiddler device present !! */
  if (!check_twiddler())
    {
      printf ("Port OK, but Twiddler data not detected.\n");
      quit (1);
    }

  

#ifdef DEBUG
  printf ("All ports opened...\n");
#endif
  
  /* OK, DONE! */
}

int unique()
{
  FILE *fd;
  char cmdline[50],pid[8],file[30];
  
  /* yup, I know this doesn't mean uniqueness but it probably does */
  /* so that's good enough for now */
  if (!(fd=fopen (TWID_FILE,"r")))
    {
      return(1);
    }
#ifdef linux
  /* let's try to clean up a stray twid.pid... */
  fscanf (fd,"%s", pid);
  fclose (fd);
  strcpy(file,"/proc/");
  strcat(file,pid);
  strcat(file,"/cmdline");
  if (!(fd=fopen(file,"r")))
    {
      unlink (TWID_FILE); /* there is no process running, erase .pid file */
      return (1);
    }
  
  fscanf (fd,"%s",cmdline);
  fclose (fd);
  if (strlen (cmdline)==0)
    return (1);
  if (strstr(cmdline,"twiddler")) /* finds a substring */
    return (0); /* twiddler is definately running! */
  return (1); 
/* twid.pid does not name a pid that is running a file called "twiddler" */
#endif
  fclose(fd);
  return (0);
}

int read_file (char *filename)
{
  int linenum,index,i,bucket=0;
  char line[MAXLINE], *map;
  FILE *user;
 
  linenum=1;
  if (!(user = fopen (filename,"r")))
    return (-1);
  while (fgets(line,MAXLINE,user))
    {
      parse_line(line,linenum);
      linenum++;
    }
  fclose (user);
  return (1);
}

int parse_line (char *line,int linenum)
{
  char cur_string[MAXLINE],tempmap[MAXLINE],key[MAXLINE],thumb[MAXLINE];
  int mapplace=0,index=0,i,t_index;
 
  /* format is :
     <thumb>+<thumb>... <four character code> = "string to translate"
     note the quotes, up to five entries (MOUSE is reserved) for thumb
     separated by a '+' */

  if (strlen(line)<2)
    return (-1);

  t_index=0;
  raise_case (line);
  sscanf(line,"%s",thumb);
  cropline (&line);
  
  do
    {
      i=get_thumb (thumb,cur_string,&t_index);
      if (!strcmp (cur_string, "CTRL"))
	index += CTRL_INDEX;
      else
	if (!strcmp(cur_string, "ALT"))
	  index += ALT_INDEX;
	else
	  if (!strcmp(cur_string, "NUM"))
	    index += NUM_INDEX;
	  else
	    if (!strcmp(cur_string, "FUNC"))
	      index += FUNC_INDEX;
	    else
	      if (!strcmp(cur_string, "SHIFT"))
		index += SHIFT_INDEX;
	      else
		if (!strcmp(cur_string, "0"));
		else
		  {
		    printf ("Error in config file, line %d: %s not a valid thumb button keyword\n",linenum,cur_string);
		    return (-1);
		  }
    } while (i);
  sscanf(line,"%s",cur_string);
  cropline (&line);
  if (strlen (cur_string)!=4)
    {
      printf ("Error in config file, line %d: Finger field %s is invalid\n",linenum,cur_string);
      return(-1);
    }
  else
    for (i=0;i<4; i++)
      {
	switch (cur_string[i])
	  {
	  case 'L':
	    index += LEFT_MASK << (2*i);
	    break;
	  case 'M': 
	    index += MID_MASK << (2*i);
	    break;
	  case 'R':
	    index += RIGHT_MASK << (2*i);
	    break;
	  case '0':
	    break;
	  default:
	  printf ("Error in config file,line %d: Finger field %s is invalid\n",linenum,cur_string);
	  return (-1);
	  }
      }
  
  sscanf(line,"%s",cur_string);
  cropline (&line);
  
  if ((strlen(cur_string)!=1) || (cur_string[0]!='='))
    {
      printf ("Error in config file, line %d: Equal sign Missing\n",linenum);
      return(-1);
    }
  /* three ways to get stuff: keywords, numbers, and strings */
  while ((line[0]==' ')||(line[0]=='\t'))
    line++;
  while (line[0]&&line[0]!='\n')
    {
      if (line[0]=='\"') 
	{ 
	  if (!get_string (&line,key,linenum)) return (-1); 
	}
      else if ((line[0]>='0') && (line[0] <='9'))
	{ 
	  if (!get_num (&line,key,linenum)) return (-1); 
	}
      else 
	{ 
	  if (!get_keyword(&line,key,linenum)) return (-1);
	}
      strcpy (tempmap+mapplace,key);
      mapplace += strlen(key);
      while ((line[0]==' ')||(line[0]=='\t'))
	line++;
    }
  table[index] = calloc (strlen(tempmap)+1,sizeof(char));
  strcpy (table[index],tempmap);
  return (index);
}


void raise_case (char *line)
{
int i,quote=0;
/* puts the string to all CAPS except what's in quotes */
for (i=0; i<(strlen(line)) ; i++)
  {
    if (!quote)
      {
	if (line[i]=='\"')
	  quote=1;
	else if  ((line[i]>='a')&&(line[i]<='z'))
	  line [i] += ('A'-'a');
      }
    else
      if (line[i]=='\"')
	quote=0;
  }
}

void cropline (char **line)
{
  /* gets rid of all whitespace, then one word (until next whitespace) */
  int flag=0;
  
  while (!(flag&&((**line==' ') || (**line=='\t'))))
    {
      if ((!flag) && ((**line != ' ') && (**line != '\t')))
	flag =1;
      (*line)++;
    }
}

int get_thumb (char *thumb,char *word,int *index)
{
  int begin=*index,i;

  while ((thumb[*index]!='\0') && (thumb[*index]!='+') &&
	 (thumb[*index]!='\t') && (thumb[*index]!=' '))
    (*index)++;
  
  for (i=0;i<10;i++)
    word[i]=0;
  for (i=begin;i<*index;i++)
    word[i-begin]=thumb[i];
  
  if (thumb[*index]=='+')
    {
      (*index)++;
      return (1);
    }
  else return (0);
}

int get_string(char** cur_string,char *key,int line)
{
  int index=0;

  (*cur_string)++; 
  while (**cur_string!='\"')
    {
      if (**cur_string=='\0') return (0);
      key[index]=**cur_string;
      index ++; (*cur_string)++;
    }
  key[index]='\0';
  (*cur_string)++;
  /* get rid of , */
  while ((**cur_string==' ')||(**cur_string=='\t'))
    (*cur_string)++;
  if (**cur_string==',')
   {
     (*cur_string)++;
     return (1);
   }
  if ((**cur_string=='\0')||(**cur_string=='\n'))
    return (1);
  printf ("Error in config file, line %d: String argument invalid\n",line);
  return(0);

}

int get_num (char** cur_string,char *key,int line)
{
  int num=0;
 
  while ((**cur_string>='0') && (**cur_string <='9'))
    {
      num *= 10;
      num += **cur_string-'0';
      (*cur_string)++;
    }
  key[0]=num;
  key[1]='\0';
  while ((**cur_string==' ')||(**cur_string=='\t'))
    (*cur_string)++;
  if (**cur_string==',')
    {
      (*cur_string)++;
      return (1);
    }
  if ((**cur_string=='\0')||(**cur_string=='\n'))
    return (1);
  printf ("Error in config file, line %d: Number argument invalid\n",line);
  return(0);
}

int get_keyword (char **cur_string, char *key,int line)
{
  int index,i;
  char word[20]="";
  index=0;

  while (((*cur_string)[index]!='\0') && ((*cur_string)[index]!=',') &&
	 ((*cur_string)[index]!='\t') && ((*cur_string)[index]!=' ') &&
	 ((*cur_string)[index]!='\n'))
    index++;
 
  strncpy (word,*cur_string,index);
  
  if (((*cur_string)[index]=='\0')||((*cur_string)[index]=='\n'))
    *cur_string+=index;
  else *cur_string+=index+1;
  while ((**cur_string==' ')||(**cur_string=='\t'))
    (*cur_string)++; 
  if (!strcmp (word,"BACKSPACE"))
    {
      key[0] = BACKSPACE_CODE;
      key[1]='\0';
      return (1);
    }
  if (!strcmp (word,"ENTER"))
    {
      key[0] = ENTER_CODE;
      key[1]='\0';
      return (1);
    }
  if (!strcmp (word, "CTRLALTDEL"))
   { 
      for (i=0;i<=4;i++) key[i]=27;
      key[5]=6;
      key[6]='\0';
      return (1);
    }
  if (!strcmp (word,"ESCAPE"))
    {
      key[0] = ESCAPE_CODE;
      key[1]='\0';
      return (1);
    }
  if (!strcmp (word,"TAB"))
    {
      key[0] = TAB_CODE;
      key[1]='\0';
      return (1);
    }
  if (!strcmp (word,"NULL"))
    {
      for (i=0;i<=4;i++) key[i]=27;
      key[5]=4;
      key[6]='\0';
      return (1);
    }
  if (!strcmp (word,"CAPS_LOCK"))
    {
      for (i=0;i<=4;i++) key[i]=27;
      key[5]=0;
      key[6]='\0';
      return (1);
    }
  if (!strcmp (word,"SCROLL_LOCK"))
    {
      for (i=0;i<=4;i++) key[i]=27;
      key[5]=2;
      key[6]='\0';
      return (1);
    }
  if (!strcmp (word,"VT"))
    {
      for (i=0;i<=4;i++) key[i]=27;
      key[5]=5;
      key[6]='\0';
      return (1);
    }
  if (!strcmp (word,"NUM_LOCK"))
    {
      for (i=0;i<=4;i++) key[i]=27;
      key[5]=1;
      key[6]='\0';
      return (1);
    }
  if (!strcmp (word,"X_TOGGLE"))
    {
      for (i=0;i<=4;i++) key[i]=27;
      key[5]=3;
      key[6]='\0';
      return (1);
    }
  
  printf ("Error in config file, line %d: %s is an Invalid Keyword\n",line,word);
  return (0);
}

void X_off (int signo)
{
  if (signal(SIGPIPE,X_off)==SIG_ERR)
    { 
      printf ("Can't get signal SIGPIPE \n");
      quit (1);
    }
  /** write to X-pipe failed! catch signal! 
    and dont die! */
  a2x_file=NULL;
#ifdef linux
  m_info.xterm_on=0;
  /* switch to VT #1 (to avoid more failures) */
  sleep(3);
  
  if (X_VT == 1)
    current_vt = 2;
  else
    current_vt = 1; /* trick twiddler int thinking we just changed
		       from 1 so it will restart */

/* I commented out the lines below because we shouldn't actually
   switch ourselves on failure. the user still has control of that
   even if the ther keys don't work */
/*  ttyfd = open ("/dev/console",O_WRONLY); 
  ioctl (ttyfd, VT_ACTIVATE, 1);	   
  close (ttyfd);
*/
#endif
}

void quit (int signo)
{
  if (signo==SIGQUIT)
    exit (0);
  unlink (TWID_FILE);
  unlink (TWID_TEMP);
  close (twidfd);
  close (ttyfd);
  free_table();
  if (a2x_file)
    pclose (a2x_file);

  fprintf (stderr,"Twiddler Driver Killed..\n");
  fflush (stderr);
  exit (1);
}

void root_error ()
{
  printf ("It's not nice to try that if you aren't root!\n");
  quit(SIGQUIT);
}

void reinit (int signo)
{
  FILE *temp;
  char text [256];
  char *filename;

  if (!(filename = getenv ("TWID_DEFAULTS")))
    {
      filename = calloc (20,sizeof(char ));
      strcpy (filename , DEFAULT_FILE);
    }
  if (temp=fopen (TWID_TEMP,"r"))
    {
      fscanf (temp,"%s",text);
      fclose (temp);
      unlink (TWID_TEMP);
      free_table();
      init_table();
      if (read_file(filename)==-1)
	{
	  fprintf (stderr,"Twiddler Keymap: Error reading default file %s !!\n",filename);
	  fprintf (stderr,"Check that file exists and try again\n");
	  quit (SIGTERM);
	}
      fprintf (stdout,"\nTwiddler Defaults restored.\n");
      fflush (stdout);
      if (strcmp(text,filename))
	{
	  if (read_file(text)==-1)
	    {
	      fprintf (stderr,"Twiddler Keymap: Error reading file %s!\n",text);
	      fprintf (stderr,"Additions will not be active...\n");
	    }
	}
    }
  else
    {
      fprintf (stderr,"Could not open message file\n");
    }
  return;
}

int check_twiddler ()
{
  char buffer[40];
  int numread,total,j,found,BAD_SEQ,first;
  long flags;
  time_t timebegin;

  found=0;
  fprintf (stderr,"Checking twiddler.. ");
  flags = fcntl (twidfd, F_GETFL);
  fcntl (twidfd, F_SETFL, flags|O_NONBLOCK);
  timebegin = time(0);
  numread=0;
  total =0;
  while ((time(0)<timebegin+10) && total<20)
    {
      numread = read (twidfd, buffer+total, 20-total);
      if (numread >0)
	total += numread;
    }
  fcntl (twidfd, F_SETFL, flags);
  fprintf (stderr,"%d/20 bytes read\n",total);
  if (total <20)
    {
      fprintf (stderr,"Could not read Twiddler data\n");
      return (0);
    }
  for (j=0;(j<6 && !found); j++)
    {
      /* find header byte */
      if (!(*(buffer+j)&0x80))
	{
	  first =j;
	  found = 1;
	}
    }
  if (!found)
    {
      fprintf (stderr,"Couldn't find 1st byte\n");
      return (0);
    }
  /* now look at the rest.. */
  BAD_SEQ=0;
  for (j=first;(j<20 && !BAD_SEQ);j++)
    {
      if ((j-first)%5 == 4)
	{
	  if (!(buffer[j]&0x80))
	    BAD_SEQ = 0x40;		    
	  /* check last byte*/
	  if ((buffer[j]&0xf0) != 0x80)
	    BAD_SEQ = 0x40;		     
	}
      else if ((j-first)%5)
	{
	  /* 1-4 */
	  if (!(buffer[j]&0x80))
	    BAD_SEQ = 0x80;		    
	}
      else if (buffer[j]&0x80)
	BAD_SEQ=0xb0;
    }
  if (BAD_SEQ)
    {
      fprintf (stderr,"Bad Twiddler data (%x)\n",BAD_SEQ);
    }
  fprintf (stderr,"Twiddler OK\n");
  fflush (stderr);
  return (1);
  
  /* found the twiddler !*/
}

void handle_mouse_activity (int index, STATE events)  
{
  int delta;
  int dx,dy,ij;
  struct keylist * temp;

  if ((index&0x30)==m_info.m_hr_ind)
    m_info.mouse_res=m_info.mid_res;
  else if ((index&0x30)==m_info.m_xhr_ind)
    m_info.mouse_res=m_info.high_res;
  else
    m_info.mouse_res=m_info.norm_res;

  filter_mouse (m_info.mouse_filter,m_info.mouse_res,&dx,&dy,
		m_info.FILTER_LEVEL);     
  

  if (!m_info.mouse_on)
    m_info.mouse_on=1;

  if (dx||dy)
    {
#ifdef XSTUFF
      if (!m_info.m_samplesize)
	{

	  if (dpy==NULL)
	    {
	      dpy = XOpenDisplay (":0.0");
	    }
	  XWarpPointer(dpy, None, None, 0, 0, 0, 0, dx, dy);
	  XFlush (dpy);
#endif
#ifndef  XSTUFF

	  if (a2x_file) fprintf (a2x_file, "%c%c%d %d%c",20,4,dx, dy,20); 
	  if (a2x_file) fflush (a2x_file);
#endif
#ifdef XSTUFF
	}
      else
	if (m_info.m_samplesize == m_info.m_samplenum)
	  {
	    if (dpy==NULL)
	      {
		dpy = XOpenDisplay (":0.0");
	      }
	    XWarpPointer(dpy, None, None, 0, 0, 0, 0, m_info.saved_dx+dx, 
			 m_info.saved_dy+dy);
	    XFlush (dpy);
	    m_info.m_samplenum=0;
	    m_info.saved_dx=0;	    
	    m_info.saved_dy=0;
	  }
	else
	  {
	    m_info.saved_dx += dx;
	    m_info.saved_dy += dy;
	    m_info.m_samplenum++;
	  }
#endif
    }

  index &= 0xff;

  /* move mouse*/
  if (!(events->flags&EVENT_ENABLE))
    return ; /* if key events not enabled, let's return */
  
  /* check to get special mouse events */

  if (events->cur_state.index > events->last_state.index)
    switch (index&0xcf)
      {
      case (LEFT_BUTTON):
	if (a2x_file) fprintf (a2x_file,"%c%c1%c",20,2,20);
	if (a2x_file) fflush(a2x_file);
	events->flags &= ~EVENT_ENABLE;
	break;
      case (MID_BUTTON):
	if (a2x_file) fprintf (a2x_file,"%c%c2%c",20,2,20);
	if (a2x_file) fflush (a2x_file);
	events->flags &= ~EVENT_ENABLE;
	break;
      case (RIGHT_BUTTON):
	if (a2x_file) fprintf (a2x_file,"%c%c3%c",20,2,20);
	if (a2x_file) fflush(a2x_file);
	events->flags &= ~EVENT_ENABLE;
	break;
      }
  
  if (events->cur_state.index < events->last_state.index)
    {
      if (events->last_state.index==m_info.m_rev_ind)
	{
	  events->flags &= ~EVENT_ENABLE;
	  m_info.mouse_reverse = 1-m_info.mouse_reverse;
	  return ;
	}

	  /* we dont have any other keys to be pressed during mouse mode,
	     but here's where normal style "programmed" key presses 
	     like mouse_reverse would go */
    }
  return;
} 

void filter_mouse (struct dxdy *filter,
		   int res_level, int *dx, int * dy,
		   int filter_size)
{
  int i,cx,cy,curdx,curdy,max,maxd;
  int tdx,tdy,delta;
  int fil_add,filter_level;
  int denominator;
  *dx = 0;
  *dy = 0;
 
  /** filter_level calc from Win 3.1 Driver **/
  /* MAX(abs(cx,cy) */
  cx = filter->cxraw;
  cy = filter->cyraw;
  if (cx<0) cx=-cx;
  if (cy<0) cy=-cy;
  if (cx>cy)
    max=cx;
  else
    max=cy;
  /* MAX(abs(dx,dy) */
  curdx = filter->dx;
  curdy = filter->dy;
  if (curdx<0) curdx=-curdx;
  if (curdy<0) curdy=-curdy;

  if (m_info.tilt_divisor == 0)
    filter_level = filter_size;
  else    
    filter_level = (max/(m_info.tilt_divisor))+m_info.tilt_constant;
  /* now do dynamic part*/
  if (filter_level - m_info.filmin<0)
    filter_level = m_info.filmin;
  else
    {
      maxd=curdx+curdy;
      if (maxd>m_info.delmax)
	maxd = m_info.delmax;
      maxd*=100;
      if (m_info.delmax !=0)
	maxd/= m_info.delmax;
      else 
	maxd=0;
      maxd = 100-maxd; /* higher deltas mean LOWER filter */
      maxd*=(filter_level - m_info.filmin);
      maxd/=100;
      filter_level = maxd + m_info.filmin;
    }
  /* ending sanity check */
  if (filter_level>filter_size) 
    filter_level = filter_size;
  if (filter_level<1)
    filter_level = 1;

  /** end filter_level calc **/

  for (i=0;i<(filter_level-1); i++)
    {
      filter=filter->prev;
    }
  for (i=0;i<(filter_level-1); i++)
    {
      
      *dx += filter->dx;
      *dy += filter->dy;
      filter= filter->next;
    }

  tdx = filter->dx;
  tdy = filter->dy;
  /* weight the current one more */
  /*	  dsq = (tdx*tdx)+(tdy*tdy); */
  
  delta = tdx+tdy - m_info.accel_thresh;
  if (delta>0)
    {
      /*	      del = sqrt_est (dsq)-m_info.accel_thresh; */
      fil_add =  (m_info.sensitivity+(delta*m_info.accel_factor));
      *dx += tdx*fil_add;
      *dy += tdy*fil_add;	      
    }
  else
    {
      fil_add = m_info.sensitivity;
      *dx += tdx*fil_add;
      *dy += tdy*fil_add;
    }

  *dx*=res_level;
  *dy*=res_level;
  /* The 10 below just helps us set the sensitivity more accurately */
  /* "16" instead of "1.6" */
  denominator = 10*((filter_level-1)+fil_add);
  /* Make sure the remainder scales correctly */
  m_info.dx_rem = (m_info.dx_rem * denominator)/m_info.old_den;
  m_info.dy_rem = (m_info.dy_rem * denominator)/m_info.old_den;
  *dx+= m_info.dx_rem; /* small movememts ad up from last time. */
  *dy+= m_info.dy_rem;


  m_info.old_den = denominator;
  m_info.dx_rem = *dx%(denominator);
  m_info.dy_rem = *dy%(denominator);
  *dx /= denominator;
  *dy /= denominator;
}
